import numpy as np
import pandas as pd

def in_silico_fold_change(concentration, fold_change_limit=9):
    ''' Take a series of known concentration, and a fold change limit, and relabel
    the concentrations such that new labels are within the fold change limit.

    Returns the given concentrations, the replaced concentration, the fold change between them,
    as well as the dictionary used to make the relabeling.
    '''
    conc_df = pd.DataFrame(concentration)
    conc_df.columns = ['concentration']
    global_candidates = conc_df.index

    # Create swappings which are consistent with fold change limit
    candidates = {}
    for e in conc_df.index:
        cc = conc_df.loc[e, 'concentration']
        ll = cc / fold_change_limit
        ul = cc * fold_change_limit
        candidates[e] = conc_df.query('{} < concentration < {}'.format(ll, ul)).index
        candidates[e] = candidates[e].drop(e)

    replacement = {}
    for e in conc_df.index:
        if e not in global_candidates:
            continue

        possible_swaps = global_candidates.intersection(candidates[e])
        # break
        if len(possible_swaps) < 1:
            replacement[e] = e
            global_candidates = global_candidates.drop(replacement[e])
        else:
            replacement[e] = np.random.choice(possible_swaps, replace=False)
            replacement[replacement[e]] = e
            global_candidates = global_candidates.drop(replacement[e])
            global_candidates = global_candidates.drop(e)

    # Make randomly swapped annotation
    shuff_concentration = concentration.copy().rename(replacement)
    concentration = concentration.sort_index()
    shuff_concentration = shuff_concentration.sort_index()

    log2_fc = np.log2(shuff_concentration / concentration)

    return concentration, shuff_concentration, log2_fc, replacement


def in_silico_conditions(expression_table, replacement):
    ''' Takes and expression table, and a dictionary for remapping gene names
    in the expression table.

    This randomly partitions the table in to two conditions, A and B.
    The B condition will have genes renamed based on the replacement dict.

    Returns the new table, and annotation about which is which.
    '''
    # Split samples in two
    n_samples = expression_table.columns.shape[0]
    shuffled_samples = np.random.choice(expression_table.columns, n_samples, replace=False)
    A_samples = shuffled_samples[:n_samples // 2]
    B_samples = shuffled_samples[n_samples // 2:]

    # Swap input abundance annotation for half of the samples
    A_table = expression_table[A_samples]
    B_table = expression_table[B_samples]
    B_table = B_table.rename(index=replacement)

    # Put everything back to a single table
    shuff_table = A_table.join(B_table)

    # Create annotation for the shuffled samples
    sample_info = pd.DataFrame({'n_genes': (shuff_table > 1.).sum(0)})
    sample_info['condition'] = \
    ['A' if s in A_samples else 'B' for s in sample_info.index]

    return shuff_table, sample_info



import warnings

import numpy as np
import pandas as pd
from tqdm import tqdm
from scipy import optimize, stats


def impulse(t, h0, h1p, h2, t1, t2, beta):
    ''' A parametric impulse function defined by two conjoined sigmoids.

    See ImpulseDE2 paper.
    '''
    return 1. / (h1p + 1) * \
           (h0 + (h1p - h0) / (1. + np.exp(-np.exp(beta) * (t - t1)))) * \
           (h2 + (h1p - h2) / (1. + np.exp( np.exp(beta) * (t - t2))))


def make_p0(run_data):
    ''' Reasonable parameter initialization for the impulse model.
    '''
    h0_init = run_data.loc[run_data.hour.argmin(), 'expr']
    peak = run_data.loc[run_data.expr.argmax()]
    h1_init = peak['expr']
    h2_init = run_data.loc[run_data.hour.argmax(), 'expr']

    t1_init = (run_data['hour'].min() + peak['hour']) / 2
    t2_init = (run_data['hour'].max() + peak['hour']) / 2
    
    beta_init = -1.
    
    set_start_values = (0., h1_init, 0., t1_init, t2_init, beta_init)
    
    return set_start_values


def make_run_data(gene, t, expression_matrix):
    return pd.DataFrame({
        'hour': t,
        'expr': expression_matrix.loc[gene]
    })


# Helpers for the optimize.leastsq function
def func(p, x, y):
    return impulse(x, *p)

def residuals(p, x, y):
    return func(p, x, y) - y


def impulse_tests(t, expression_matrix, maxfev=50):
    ''' Least squares version of the ImpulseDE2 test

        t : A Series or vector with time values
        expresion_matrix: Assume columns are genes and rows are sampels

        maxfev: Maximum number of function evaluations per gene, higher numbers 
                give better accuracy at the cost of speed. Defualt is 50.
    '''
    t = np.array(t)
    n = expression_matrix.shape[1]

    gene_params = pd.DataFrame(index=expression_matrix.index,
                               columns=['h0', 'h1', 'h2', 't1', 't2', 'beta', \
                                        'llr', 'pval', 'res_alt', 'res_null'])

    with warnings.catch_warnings():
        warnings.simplefilter('ignore', category=RuntimeWarning)
        for gene in tqdm(expression_matrix.index):
            run_data = make_run_data(gene, t, expression_matrix)

            opt_res = optimize.leastsq(residuals,
                                       make_p0(run_data),
                                       args=(t, run_data.expr.values),
                                       maxfev=maxfev)

            gene_params.loc[gene, ['h0', 'h1', 'h2', 't1', 't2', 'beta']] = opt_res[0]

            res_alt = np.sum(np.square(residuals(opt_res[0], t, run_data.expr.values)))
            gene_params.loc[gene, 'res_alt'] = res_alt
            ll_alt = -n / 2. * np.log(2 * np.pi) - n / 2. * np.log(res_alt / n) - n / 2.
            
            res_null = np.sum(np.square(run_data.expr.values.mean() - run_data.expr.values))
            gene_params.loc[gene, 'res_null'] = res_null
            ll_null = -n / 2. * np.log(2 * np.pi) - n / 2. * np.log(res_null / n) - n / 2.

            gene_params.loc[gene, 'llr'] = ll_alt - ll_null

    gene_params['llr'] =  gene_params['llr'].replace(-np.inf, 0).fillna(0)
    gene_params['pval'] = stats.chi2.sf(2 * gene_params['llr'], df=5)

    return gene_params




import numpy as np
import pandas as pd
import patsy
from scipy import stats


def lr_tests(sample_info, expression_matrix, alt_model, null_model='~ 1', rcond=-1):
    ''' Compare alt_model and null_model by a Likelihood Ratio Test for every
    gene in the expression_matrix.
    
    Assumes columns are samples, and rows are genes.
    
    OLS Regression assumes data is variance stabilized. E.g. log scaled.
    
    Returns a DataFrame with alt_model weights and p-values.
    '''
    alt_design  = patsy.dmatrix(alt_model, sample_info, return_type='dataframe')
    null_design = patsy.dmatrix(null_model, sample_info, return_type='dataframe')
    
    beta_alt,  res_alt,  rank_alt,  s_alt = np.linalg.lstsq(alt_design, expression_matrix.T, rcond=rcond)
    beta_null, res_null, rank_null, s_null = np.linalg.lstsq(null_design, expression_matrix.T, rcond=rcond)
    
    results = pd.DataFrame(beta_alt.T, columns=alt_design.columns, index=expression_matrix.index)
    
    n = expression_matrix.shape[1]
    ll_alt  = -n / 2. * np.log(2 * np.pi) - n / 2. * np.ma.log(res_alt  / n) - n / 2.
    ll_null = -n / 2. * np.log(2 * np.pi) - n / 2. * np.ma.log(res_null / n) - n / 2.
    
    llr = ll_alt - ll_null
    
    pval = stats.chi2.sf(2 * llr, df=beta_alt.shape[0] - beta_null.shape[0])
    pval = np.ma.MaskedArray(pval, mask=llr.mask).filled(1.)
    
    results['pval'] = pval
    results['qval'] = (results['pval'] * n).clip_upper(1.)
    
    return results


def regress_out(sample_info, expression_matrix, covariate_formula, design_formula='1', rcond=-1):
    ''' Implementation of limma's removeBatchEffect function
    '''
    # Ensure intercept is not part of covariates
    covariate_formula += ' - 1'

    covariate_matrix = patsy.dmatrix(covariate_formula, sample_info)
    design_matrix = patsy.dmatrix(design_formula, sample_info)

    design_batch = np.hstack((design_matrix, covariate_matrix))

    coefficients, res, rank, s = np.linalg.lstsq(design_batch, expression_matrix.T, rcond=rcond)
    beta = coefficients[design_matrix.shape[1]:]
    regressed = expression_matrix - covariate_matrix.dot(beta).T

    return regressed


def stabilize(expression_matrix):
    ''' Use Anscombes approximation to variance stabilize Negative Binomial data

    See https://f1000research.com/posters/4-1041 for motivation.

    Assumes columns are samples, and rows are genes
    '''
    from scipy import optimize
    phi_hat, _ = optimize.curve_fit(lambda mu, phi: mu + phi * mu ** 2, expression_matrix.mean(1), expression_matrix.var(1))

    return np.log(expression_matrix + 1. / (2 * phi_hat[0]))

def anscombe(exppression_matrix):
    ''' Use Anscombes approximation to variance stabilize Negative Binomial data

    See https://f1000research.com/posters/4-1041 for motivation.

    Assumes columns are samples, and rows are genes
    '''
    return stabilize(expression_matrix)


def vst(expression_matrix):
    ''' A VST derived from assumption of a global NB phi parameter for all genes.

    Var(mu) = mu + phi * mu^2

    Defined by symbolic integral as described in http://www.bioconductor.org/packages//2.13/bioc/vignettes/DESeq/inst/doc/vst.pdf

    Unlike the `stabilize` function, results here will be non-negative. This also
    assumes phi > 0.

    Assumes columns are samples, and rows are genes
    '''
    from scipy import optimize
    v = lambda mu, phi: mu + phi * mu ** 2
    phi_hat, _ = optimize.curve_fit(v, expression_matrix.mean(1), expression_matrix.var(1))

    return 2 * np.arcsinh(np.sqrt(phi_hat[0] * expression_matrix)) / np.sqrt(phi_hat[0])



from setuptools import setup, find_packages

def readme():
    with open('README.md') as f:
        return f.read()

setup(
        name='NaiveDE',
        version='1.2.0',
        description='The most trivial DE test based on likelihood ratio tests',
        long_description=readme(),
        packages=find_packages(),
        install_requires=['numpy', 'scipy', 'pandas', 'tqdm'],
        author='Valentine Svensson',
        author_email='valentine@nxn.se',
        license='MIT'
    )
